# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'qtCreatorFiles/ui_annotator_rhm.ui'
#
# Created by: PyQt5 UI code generator 5.14.2
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_annotator_rhm(object):
    def setupUi(self, annotator_rhm):
        annotator_rhm.setObjectName("annotator_rhm")
        annotator_rhm.setWindowModality(QtCore.Qt.WindowModal)
        annotator_rhm.resize(862, 670)
        self.horizontalLayout = QtWidgets.QHBoxLayout(annotator_rhm)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.hideableBox = QtWidgets.QFrame(annotator_rhm)
        self.hideableBox.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.hideableBox.setFrameShadow(QtWidgets.QFrame.Raised)
        self.hideableBox.setObjectName("hideableBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.hideableBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.frame1 = QtWidgets.QFrame(self.hideableBox)
        self.frame1.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame1.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame1.setObjectName("frame1")
        self.modeLayout = QtWidgets.QHBoxLayout(self.frame1)
        self.modeLayout.setContentsMargins(0, 0, 0, 0)
        self.modeLayout.setObjectName("modeLayout")
        self.hamMenuButton = QtWidgets.QToolButton(self.frame1)
        self.hamMenuButton.setText("☰")
        self.hamMenuButton.setObjectName("hamMenuButton")
        self.modeLayout.addWidget(self.hamMenuButton)
        self.finishedButton = QtWidgets.QPushButton(self.frame1)
        self.finishedButton.setWhatsThis("")
        self.finishedButton.setObjectName("finishedButton")
        self.modeLayout.addWidget(self.finishedButton)
        self.finishNoRelaunchButton = QtWidgets.QPushButton(self.frame1)
        self.finishNoRelaunchButton.setObjectName("finishNoRelaunchButton")
        self.modeLayout.addWidget(self.finishNoRelaunchButton)
        self.cancelButton = QtWidgets.QPushButton(self.frame1)
        self.cancelButton.setObjectName("cancelButton")
        self.modeLayout.addWidget(self.cancelButton)
        self.verticalLayout.addWidget(self.frame1)
        self.frame2 = QtWidgets.QFrame(self.hideableBox)
        self.frame2.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame2.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame2.setObjectName("frame2")
        self.buttonsLayout = QtWidgets.QHBoxLayout(self.frame2)
        self.buttonsLayout.setContentsMargins(0, 0, 0, 0)
        self.buttonsLayout.setSpacing(3)
        self.buttonsLayout.setObjectName("buttonsLayout")
        self.markLabel = QtWidgets.QLabel(self.frame2)
        self.markLabel.setObjectName("markLabel")
        self.buttonsLayout.addWidget(self.markLabel)
        self.zoomCB = QtWidgets.QComboBox(self.frame2)
        self.zoomCB.setObjectName("zoomCB")
        self.buttonsLayout.addWidget(self.zoomCB)
        self.verticalLayout.addWidget(self.frame2)
        self.frameTools = QtWidgets.QFrame(self.hideableBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frameTools.sizePolicy().hasHeightForWidth())
        self.frameTools.setSizePolicy(sizePolicy)
        self.frameTools.setObjectName("frameTools")
        self.toolLayout = QtWidgets.QGridLayout(self.frameTools)
        self.toolLayout.setContentsMargins(0, 0, 0, 0)
        self.toolLayout.setSpacing(3)
        self.toolLayout.setObjectName("toolLayout")
        self.penButton = QtWidgets.QToolButton(self.frameTools)
        self.penButton.setMinimumSize(QtCore.QSize(45, 0))
        self.penButton.setCheckable(True)
        self.penButton.setAutoExclusive(True)
        self.penButton.setObjectName("penButton")
        self.toolLayout.addWidget(self.penButton, 0, 4, 1, 1)
        self.textButton = QtWidgets.QToolButton(self.frameTools)
        self.textButton.setMinimumSize(QtCore.QSize(45, 0))
        self.textButton.setCheckable(True)
        self.textButton.setAutoExclusive(True)
        self.textButton.setObjectName("textButton")
        self.toolLayout.addWidget(self.textButton, 1, 4, 1, 1)
        self.commentDownButton = QtWidgets.QToolButton(self.frameTools)
        self.commentDownButton.setMinimumSize(QtCore.QSize(45, 0))
        self.commentDownButton.setToolTipDuration(-1)
        self.commentDownButton.setObjectName("commentDownButton")
        self.toolLayout.addWidget(self.commentDownButton, 2, 3, 1, 1)
        self.commentUpButton = QtWidgets.QToolButton(self.frameTools)
        self.commentUpButton.setMinimumSize(QtCore.QSize(45, 0))
        self.commentUpButton.setToolTipDuration(-1)
        self.commentUpButton.setObjectName("commentUpButton")
        self.toolLayout.addWidget(self.commentUpButton, 0, 3, 1, 1)
        self.panButton = QtWidgets.QToolButton(self.frameTools)
        self.panButton.setMinimumSize(QtCore.QSize(45, 0))
        self.panButton.setCheckable(True)
        self.panButton.setAutoExclusive(True)
        self.panButton.setObjectName("panButton")
        self.toolLayout.addWidget(self.panButton, 0, 0, 1, 1)
        self.lineButton = QtWidgets.QToolButton(self.frameTools)
        self.lineButton.setMinimumSize(QtCore.QSize(45, 0))
        self.lineButton.setCheckable(True)
        self.lineButton.setAutoExclusive(True)
        self.lineButton.setObjectName("lineButton")
        self.toolLayout.addWidget(self.lineButton, 2, 4, 1, 1)
        self.tickButton = QtWidgets.QToolButton(self.frameTools)
        self.tickButton.setMinimumSize(QtCore.QSize(45, 0))
        self.tickButton.setToolTipDuration(-1)
        self.tickButton.setCheckable(True)
        self.tickButton.setAutoExclusive(True)
        self.tickButton.setObjectName("tickButton")
        self.toolLayout.addWidget(self.tickButton, 1, 2, 1, 1)
        self.moveButton = QtWidgets.QToolButton(self.frameTools)
        self.moveButton.setMinimumSize(QtCore.QSize(45, 0))
        self.moveButton.setCheckable(True)
        self.moveButton.setAutoExclusive(True)
        self.moveButton.setObjectName("moveButton")
        self.toolLayout.addWidget(self.moveButton, 2, 0, 1, 1)
        self.redoButton = QtWidgets.QToolButton(self.frameTools)
        self.redoButton.setMinimumSize(QtCore.QSize(45, 0))
        self.redoButton.setObjectName("redoButton")
        self.toolLayout.addWidget(self.redoButton, 0, 1, 1, 1)
        self.zoomButton = QtWidgets.QToolButton(self.frameTools)
        self.zoomButton.setMinimumSize(QtCore.QSize(45, 0))
        self.zoomButton.setCheckable(True)
        self.zoomButton.setAutoExclusive(True)
        self.zoomButton.setObjectName("zoomButton")
        self.toolLayout.addWidget(self.zoomButton, 1, 0, 1, 1)
        self.deleteButton = QtWidgets.QToolButton(self.frameTools)
        self.deleteButton.setMinimumSize(QtCore.QSize(45, 0))
        self.deleteButton.setCheckable(True)
        self.deleteButton.setAutoExclusive(True)
        self.deleteButton.setObjectName("deleteButton")
        self.toolLayout.addWidget(self.deleteButton, 2, 1, 1, 1)
        self.commentButton = QtWidgets.QToolButton(self.frameTools)
        self.commentButton.setMinimumSize(QtCore.QSize(45, 0))
        self.commentButton.setToolTipDuration(-1)
        self.commentButton.setCheckable(True)
        self.commentButton.setAutoExclusive(True)
        self.commentButton.setObjectName("commentButton")
        self.toolLayout.addWidget(self.commentButton, 1, 3, 1, 1)
        self.undoButton = QtWidgets.QToolButton(self.frameTools)
        self.undoButton.setMinimumSize(QtCore.QSize(45, 0))
        self.undoButton.setObjectName("undoButton")
        self.toolLayout.addWidget(self.undoButton, 1, 1, 1, 1)
        self.crossButton = QtWidgets.QToolButton(self.frameTools)
        self.crossButton.setMinimumSize(QtCore.QSize(45, 0))
        self.crossButton.setCheckable(True)
        self.crossButton.setAutoExclusive(True)
        self.crossButton.setObjectName("crossButton")
        self.toolLayout.addWidget(self.crossButton, 0, 2, 1, 1)
        self.boxButton = QtWidgets.QToolButton(self.frameTools)
        self.boxButton.setMinimumSize(QtCore.QSize(45, 0))
        self.boxButton.setCheckable(True)
        self.boxButton.setAutoExclusive(True)
        self.boxButton.setObjectName("boxButton")
        self.toolLayout.addWidget(self.boxButton, 2, 2, 1, 1)
        self.verticalLayout.addWidget(self.frameTools, 0, QtCore.Qt.AlignHCenter)
        self.frame4 = QtWidgets.QFrame(self.hideableBox)
        self.frame4.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame4.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame4.setObjectName("frame4")
        self.ebLayout = QtWidgets.QHBoxLayout(self.frame4)
        self.ebLayout.setContentsMargins(0, 0, 0, 0)
        self.ebLayout.setSpacing(3)
        self.ebLayout.setObjectName("ebLayout")
        self.modeLabel = QtWidgets.QLabel(self.frame4)
        self.modeLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.modeLabel.setObjectName("modeLabel")
        self.ebLayout.addWidget(self.modeLabel)
        self.verticalLayout.addWidget(self.frame4)
        self.frame5_markgrid = QtWidgets.QFrame(self.hideableBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame5_markgrid.sizePolicy().hasHeightForWidth())
        self.frame5_markgrid.setSizePolicy(sizePolicy)
        self.frame5_markgrid.setObjectName("frame5_markgrid")
        self.container_markgrid = QtWidgets.QHBoxLayout(self.frame5_markgrid)
        self.container_markgrid.setContentsMargins(0, -1, 0, -1)
        self.container_markgrid.setObjectName("container_markgrid")
        self.verticalLayout.addWidget(self.frame5_markgrid)
        self.frame6 = QtWidgets.QFrame(self.hideableBox)
        self.frame6.setObjectName("frame6")
        self.hboxlayout = QtWidgets.QHBoxLayout(self.frame6)
        self.hboxlayout.setContentsMargins(0, 0, 0, 0)
        self.hboxlayout.setObjectName("hboxlayout")
        self.noAnswerButton = QtWidgets.QPushButton(self.frame6)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(2)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.noAnswerButton.sizePolicy().hasHeightForWidth())
        self.noAnswerButton.setSizePolicy(sizePolicy)
        self.noAnswerButton.setObjectName("noAnswerButton")
        self.hboxlayout.addWidget(self.noAnswerButton)
        spacerItem = QtWidgets.QSpacerItem(6, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.hboxlayout.addItem(spacerItem)
        self.rearrangePagesButton = QtWidgets.QPushButton(self.frame6)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(3)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.rearrangePagesButton.sizePolicy().hasHeightForWidth())
        self.rearrangePagesButton.setSizePolicy(sizePolicy)
        self.rearrangePagesButton.setObjectName("rearrangePagesButton")
        self.hboxlayout.addWidget(self.rearrangePagesButton)
        self.verticalLayout.addWidget(self.frame6)
        self.frame7_comment = QtWidgets.QFrame(self.hideableBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(2)
        sizePolicy.setHeightForWidth(self.frame7_comment.sizePolicy().hasHeightForWidth())
        self.frame7_comment.setSizePolicy(sizePolicy)
        self.frame7_comment.setObjectName("frame7_comment")
        self.container_commentwidget = QtWidgets.QVBoxLayout(self.frame7_comment)
        self.container_commentwidget.setContentsMargins(0, -1, 0, 0)
        self.container_commentwidget.setObjectName("container_commentwidget")
        self.verticalLayout.addWidget(self.frame7_comment)
        self.horizontalLayout.addWidget(self.hideableBox)
        self.revealBox0 = QtWidgets.QFrame(annotator_rhm)
        self.revealBox0.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.revealBox0.setFrameShadow(QtWidgets.QFrame.Raised)
        self.revealBox0.setObjectName("revealBox0")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.revealBox0)
        self.verticalLayout_2.setSpacing(2)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.revealBox1 = QtWidgets.QFrame(self.revealBox0)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.revealBox1.sizePolicy().hasHeightForWidth())
        self.revealBox1.setSizePolicy(sizePolicy)
        self.revealBox1.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.revealBox1.setFrameShadow(QtWidgets.QFrame.Raised)
        self.revealBox1.setObjectName("revealBox1")
        self.revealLayout = QtWidgets.QGridLayout(self.revealBox1)
        self.revealLayout.setContentsMargins(0, 0, 0, 0)
        self.revealLayout.setSpacing(3)
        self.revealLayout.setObjectName("revealLayout")
        self.deltaButton = QtWidgets.QToolButton(self.revealBox1)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.deltaButton.sizePolicy().hasHeightForWidth())
        self.deltaButton.setSizePolicy(sizePolicy)
        self.deltaButton.setMinimumSize(QtCore.QSize(45, 0))
        self.deltaButton.setCheckable(True)
        self.deltaButton.setAutoExclusive(True)
        self.deltaButton.setObjectName("deltaButton")
        self.revealLayout.addWidget(self.deltaButton, 1, 0, 1, 1)
        self.verticalLayout_2.addWidget(self.revealBox1, 0, QtCore.Qt.AlignTop)
        self.revealBox2 = QtWidgets.QFrame(self.revealBox0)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.revealBox2.sizePolicy().hasHeightForWidth())
        self.revealBox2.setSizePolicy(sizePolicy)
        self.revealBox2.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.revealBox2.setFrameShadow(QtWidgets.QFrame.Raised)
        self.revealBox2.setObjectName("revealBox2")
        self.revealLayout2 = QtWidgets.QVBoxLayout(self.revealBox2)
        self.revealLayout2.setContentsMargins(3, 3, 3, 3)
        self.revealLayout2.setSpacing(3)
        self.revealLayout2.setObjectName("revealLayout2")
        self.verticalLayout_2.addWidget(self.revealBox2, 0, QtCore.Qt.AlignHCenter|QtCore.Qt.AlignBottom)
        self.horizontalLayout.addWidget(self.revealBox0)
        self.pageFrame = QtWidgets.QFrame(annotator_rhm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(2)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pageFrame.sizePolicy().hasHeightForWidth())
        self.pageFrame.setSizePolicy(sizePolicy)
        self.pageFrame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.pageFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.pageFrame.setObjectName("pageFrame")
        self.pageFrameGrid = QtWidgets.QGridLayout(self.pageFrame)
        self.pageFrameGrid.setContentsMargins(0, 0, 0, 0)
        self.pageFrameGrid.setObjectName("pageFrameGrid")
        self.horizontalLayout.addWidget(self.pageFrame)

        self.retranslateUi(annotator_rhm)
        QtCore.QMetaObject.connectSlotsByName(annotator_rhm)

    def retranslateUi(self, annotator_rhm):
        _translate = QtCore.QCoreApplication.translate
        annotator_rhm.setWindowTitle(_translate("annotator_rhm", "Annotate paper"))
        self.finishedButton.setToolTip(_translate("annotator_rhm", "Save and move to the next paper"))
        self.finishedButton.setText(_translate("annotator_rhm", "Next"))
        self.finishNoRelaunchButton.setToolTip(_translate("annotator_rhm", "Save and return to marker window"))
        self.finishNoRelaunchButton.setText(_translate("annotator_rhm", "Done"))
        self.cancelButton.setToolTip(_translate("annotator_rhm", "Cancel the current annotations and return to marker window"))
        self.cancelButton.setText(_translate("annotator_rhm", "&Cancel"))
        self.markLabel.setText(_translate("annotator_rhm", "kk out of nn"))
        self.penButton.setToolTip(_translate("annotator_rhm", "press t"))
        self.penButton.setText(_translate("annotator_rhm", "..."))
        self.textButton.setToolTip(_translate("annotator_rhm", "press g"))
        self.textButton.setText(_translate("annotator_rhm", "..."))
        self.commentDownButton.setToolTip(_translate("annotator_rhm", "press v"))
        self.commentDownButton.setText(_translate("annotator_rhm", "..."))
        self.commentUpButton.setToolTip(_translate("annotator_rhm", "press r"))
        self.commentUpButton.setText(_translate("annotator_rhm", "..."))
        self.panButton.setToolTip(_translate("annotator_rhm", "press q"))
        self.panButton.setText(_translate("annotator_rhm", "..."))
        self.lineButton.setToolTip(_translate("annotator_rhm", "press b"))
        self.lineButton.setText(_translate("annotator_rhm", "..."))
        self.tickButton.setToolTip(_translate("annotator_rhm", "press d"))
        self.tickButton.setText(_translate("annotator_rhm", "..."))
        self.moveButton.setToolTip(_translate("annotator_rhm", "press z"))
        self.moveButton.setText(_translate("annotator_rhm", "..."))
        self.redoButton.setToolTip(_translate("annotator_rhm", "press w"))
        self.redoButton.setText(_translate("annotator_rhm", "..."))
        self.zoomButton.setToolTip(_translate("annotator_rhm", "press a"))
        self.zoomButton.setText(_translate("annotator_rhm", "..."))
        self.deleteButton.setToolTip(_translate("annotator_rhm", "press x"))
        self.deleteButton.setText(_translate("annotator_rhm", "..."))
        self.commentButton.setToolTip(_translate("annotator_rhm", "press f"))
        self.commentButton.setText(_translate("annotator_rhm", "..."))
        self.undoButton.setToolTip(_translate("annotator_rhm", "press s"))
        self.undoButton.setText(_translate("annotator_rhm", "..."))
        self.crossButton.setToolTip(_translate("annotator_rhm", "press e"))
        self.crossButton.setText(_translate("annotator_rhm", "..."))
        self.boxButton.setToolTip(_translate("annotator_rhm", "press c"))
        self.boxButton.setText(_translate("annotator_rhm", "..."))
        self.modeLabel.setText(_translate("annotator_rhm", "move"))
        self.noAnswerButton.setText(_translate("annotator_rhm", "No answer"))
        self.rearrangePagesButton.setToolTip(_translate("annotator_rhm", "<html><head/><body><p>Find a missing page, discard irrelevant pages, rotate pages, etc</p></body></html>"))
        self.rearrangePagesButton.setText(_translate("annotator_rhm", "Adjust pages"))
        self.deltaButton.setText(_translate("annotator_rhm", "..."))
